'''
Contains examples of standard likelihood functions for usage in the package.
'''
import numpy as np

def isotropic_gaussian(q):
    '''
    An isotropic Gaussian likelihood function.

    Parameters
    ----------
    q : numpy array
        Position parameter


    Returns
    -------
    r : dictionary
        returns the log likelihood under the ``llh`` key, and the gradient under the ``grad`` key.
    '''

    llh = -np.sum(q * q) / 2
    grad = -q

    return {"llh": llh, "grad": grad}


def ica( q, data, idxs=None, alpha=1 ):
    '''
    Independent Component Analysis likelihood with a Gaussian prior.

    Parameters
    ----------
    q : numpy array
        An (N,1) numpy array of the position parameter
    data : numpy array
        A (N,d) array, where the d datapoints have dimensionality N.
    idxs : list or iterable, optional
        A list of data indexes to use in the calculation, default is to use all of them
    alpha : float, optional
        The variance of the Gaussian prior, default 1.


    Returns
    -------
    r : dictionary
        returns the log likelihood under the ``llh`` key, the gradient under the ``grad`` key and the gradients for each data point are given in ``grad_data``.
    '''

    if (idxs is None):
        X = data
    else:
        X = data[:,idxs]

    # Prior

    Vprior = -0.5*np.sum(q**2) / alpha
    Fprior = -q / alpha

    # LLH

    D,N = data.shape

    W = np.reshape( q , (D,D) ).T
    Winv = np.linalg.inv(W)
    Wdet = np.linalg.det(W)

    WX2 = -0.5 * np.dot( W, X )

    em2x = np.exp(2*WX2 )
    tanhx = (1.0 - em2x) / (1.0 + em2x )
    logcoshx = np.log1p( em2x ) - WX2 - 0.69314718

    V = N * np.log( np.abs( Wdet ) ) - 2*np.sum( logcoshx )

    F = -np.einsum('ij,kj->kij',tanhx,X).reshape( (D*D,-1 ) )
    F = F + np.reshape( Winv.flatten(), q.shape )

    TotalV = V + Vprior
    TotalF = (N * np.mean(F, 1, keepdims=True) + Fprior)

    return {"llh": TotalV  , "grad": TotalF, "grad_data": F}


def blr(q, data, t, idxs=None, alpha=100 ):
    '''
    Bayesian Logistic Regression with a Gaussian prior.

    Parameters
    ----------
    q : numpy array
        A (N,1) array of the N-dimensional position parameter
    data : numpy array
        A (N,d) array, where the d datapoints have dimensionality N.
    t : numpy array
        A (1,d) binary array of indicator values
    idxs : list or iterable, optional
        A list of indexes to use in the BLR calculation
    alpha : float, optional
        The variance of the Gaussian prior, default 100.


    Returns
    -------
    r : dictionary
        returns the log likelihood under the ``llh`` key, the gradient under the ``grad`` key and the gradients for each data point are given in ``grad_data``.
    '''

    Ndata = data.shape[1]

    if idxs is None:
        idxs = np.arange(Ndata)

    X = data[:, idxs].T
    t = t[:, idxs].T

    # Prior
    Vprior = -0.5 * np.sum(q ** 2) / alpha
    Fprior = -q / alpha

    # Posterior
    tv = np.dot(X, q)
    exptv = np.exp(-tv)

    VV = tv * t - np.log(1 + exptv) - tv

    #V = np.sum(tv * t) - np.sum(np.log(1 + exptv)) - np.sum(tv)
    F = X * (t - 1.0 / (1.0 + exptv))

    TotalV = (np.sum(VV) + Vprior)

    TotalF = (Ndata * np.mean(F, 0, keepdims=True) + Fprior.T).T

    return {"llh": TotalV, "llh_data":VV , "grad": TotalF, "grad_data": F.T}



"""
Euler-Maruyama (SGLD)
=====================

The classic Euler-Maruyama scheme, also known as Stochastic Gradient Langevin Dynamics (SGLD) when a stochastic approximation is used as the gradient.

This scheme uses the gradient of the posterior to move through the configuration space, without Metropolis correction. It produces a biased trajectory (though convergent as the learning rate decreases), but is a good `first-choice` in high dimension for general distributions.

Usage
^^^^^

Set ``algo="sgld"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The (potentially approximate) gradient of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- This scheme takes no additional parameters.

References
^^^^^^^^^^

This scheme is extensively studied, it should be featured heavily in any good statistics textbook.

- `Numerical Solution of Stochastic Differential Equations by Eckhard Platen and Peter Kloeden, Springer (1992)`
- `Welling, Max, and Yee W. Teh. "Bayesian learning via stochastic gradient Langevin dynamics." Proceedings of the 28th international conference on machine learning (ICML-11). 2011` `<https://www.ics.uci.edu/~welling/publications/papers/stoclangevin_v6.pdf>`__

"""
from .algorithm import Algorithm


class SGLD(Algorithm):
    def step(self, q):

        q = q + self.h * self.f + self.sq2h * self.np.random.randn(*q.shape)

        fres = self.force(q)
        self.v, self.f = fres.get("llh"), fres.get("grad")

        return q



class Algorithm:
    def __init__(self, np, ic, h, force, params):

        self.np = np
        self.h = h
        self.sq2h = np.sqrt(2 * h)
        self.sqh2 = np.sqrt(h / 2)
        self.h2 = h / 2
        self.force = force
        self.acc = 0

        fres = self.force(ic)
        self.v, self.f, self.ff = (
            fres.get("llh"),
            fres.get("grad"),
            fres.get("grad_data"),
        )

        self.p = np.random.randn(*ic.shape)
        self.xi = np.random.randn(1)

    def clear(self, q):
        self.acc = 0

        fres = self.force(q)
        self.v, self.f, self.ff = (
            fres.get("llh"),
            fres.get("grad"),
            fres.get("grad_data"),
        )
        pass

    def step(self, q):
        pass



"""
BAOAB
========================

A biased Langevin sampling routine that incorporates momentum to expedite sampling over barriers. A damping parameter can be chosen to customize the mixing with the heat bath.
In the limit of infinite damping this method becomes the LM scheme. Similarly, the BAOAB method is exact for Gaussian distributions.


Usage
^^^^^

Set ``algo="baoab"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The gradient of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- ``g`` : positive float
    (Default 1.0) The isotropic friction constant (`gamma`) used in the Langevin dynamics.

References
^^^^^^^^^^

- `Molecular Dynamics: With Deterministic and Stochastic Numerical Methods by Benedict Leimkuhler and Charles Matthews, Springer (2017)`

"""
from .algorithm import Algorithm


class BAOAB(Algorithm):
    def __init__(self, np, ic, h, force, params):
        super().__init__(np, ic, h, force, params)

        self.g = params.get("g", 1)

    def step(self, q):

        h = self.h
        h2 = self.h2

        self.p = self.p + h2 * self.f
        q = q + h2 * self.p

        c1 = self.np.exp(-self.g * h)
        c3 = self.np.sqrt(1 - c1 * c1)
        self.p = self.p * c1 + c3 * self.np.random.randn(*self.p.shape)

        q = q + h2 * self.p

        fres = self.force(q)
        self.v, self.f = fres.get("llh"), fres.get("grad")

        self.p = self.p + h2 * self.f

        return q



"""
Stochastic Gradient HMC
========================

A biased scheme for use with stochastic gradients. Damps the dynamics to compensate for additional noise coming from a stochastic gradient.

The covariance of the stochastic gradient term, denoted `B`, is calculated at every step using a user-supplied function that takes as input the current value of `B` and the `grad_data` terms, if supplied.
For reasons of ergodicity (i.e. to ensure proper mixing behavior in the method) a matrix `C>=B*h/2` must be given, where `h` is the learning rate or step size. In this implementation we use scale the identity matrix scaled by a parameter ``g`` as the `C` matrix.
The large damping parameter required can mean that this scheme can easily become unstable and slow to converge if the gradient noise is large.

Usage
^^^^^

Set ``algo="sghmc"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The gradient of the log posterior.
- ``grad_data`` :: The gradient of the log likelihood with respect to the current position, itemized over the data batch. Expects a `(N,D)` array, where the position space is `N` dimensional and the batch size is `D`.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- ``g`` : positive float
    (Default 1.0) The damping parameter used in the Langevin dynamics. Must be large enough to dominate the gradient noise.
- ``auto_friction`` : bool
    (Default True) Automatically increase the friction ``g`` if the target covariance matrix is not positive-definite.
- ``B`` : numpy array
    (Default 0) The initial value to use as the damping matrix for the stochastic gradient.
- ``grad_cov`` : function
    A function estimating the covariance of the stochastic gradient, with signature `grad_cov(B, grad_data)`.
    The current covariance estimate `B` and the `grad_data` term are supplied to the function, and it outputs an `(N,N)` numpy array as the new `B` matrix.
    If a function is not specified, the default behavior is to use the last estimate of ``B`` without any change.

References
^^^^^^^^^^

- `Chen, Tianqi, Emily Fox, and Carlos Guestrin. "Stochastic gradient hamiltonian monte carlo." International conference on machine learning. (2014)`

"""
from .algorithm import Algorithm


class SGHMC(Algorithm):
    def __init__(self, np, ic, h, force, params):
        super().__init__(np, ic, h, force, params)

        self.g = params.get("g", 1)
        self.auto_friction = params.get('auto_friction', True)
        self.grad_cov = params.get('grad_cov', self.zero_cov)
        self.B = params.get('B', 0*self.np.eye( len(ic) ) )

        self.C = self.np.eye( len(ic) ) * self.g

    def step(self, q):

        h = self.h
        h2 = self.h2

        q = q + h * self.p

        fres = self.force(q)
        self.v, self.f, self.ff = fres.get("llh"), fres.get("grad"), fres.get("grad_data")

        self.B = self.grad_cov(B=self.B, grad_data=self.ff)

        if (self.auto_friction):
            success = False
            for r in range(20):
                try:
                    CminusB = self.C - self.B*h2
                    sqCminusB = self.np.linalg.cholesky( CminusB )
                    success = True
                    break
                except self.np.linalg.LinAlgError:
                    self.g *= 1.5
                    self.C *= 1.5
            if (not success):
                raise self.np.linalg.LinAlgError
        else:
            CminusB = self.C - self.B*h2
            sqCminusB = self.np.linalg.cholesky( CminusB )

        self.p = self.p + h * self.f - h * self.np.dot(self.C, self.p)
        self.p = self.p + self.sq2h*self.np.dot(sqCminusB , self.np.random.randn(*self.p.shape))

        return q

    def zero_cov(self,B,grad_data):

        return B



"""
BADODAB
========================

A scheme that adaptively learns a constant damping term required to correct the distribution when using stochastic gradients.

Usage
^^^^^

Set ``algo="badodab"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The gradient of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- ``g`` : positive float
    (Default 1.0) The damping parameter used in the Langevin dynamics.
- ``mu`` : positive float
    (Default 1.0) The precision (reciprocal variance) of the auxillary variables distribution.

References
^^^^^^^^^^

- `Leimkuhler, Benedict, and Xiaocheng Shang. "Adaptive thermostats for noisy gradient systems." SIAM Journal on Scientific Computing 38.2 (2016)`

"""
from .algorithm import Algorithm


class BADODAB(Algorithm):
    def __init__(self, np, ic, h, force, params):
        super().__init__(np, ic, h, force, params)

        self.g = params.get("g", 1)
        self.mu = params.get("mu", 1)

    def step(self, q):

        h = self.h
        h2 = self.h2

        self.p = self.p + h2 * self.f
        q = q + h2 * self.p

        self.xi = self.xi + h2 * (self.np.sum(self.p * self.p) - self.p.size) / self.mu

        c1 = self.np.exp(-self.xi * h)
        c3 = self.np.sqrt(self.np.abs(self.g * (1 - c1 * c1) / (self.xi)))
        self.p = self.p * c1 + c3 * self.np.random.randn(*self.p.shape)

        self.xi = self.xi + h2 * (self.np.sum(self.p * self.p) - self.p.size) / self.mu

        q = q + h2 * self.p

        fres = self.force(q)
        self.v, self.f = fres.get("llh"), fres.get("grad")

        self.p = self.p + h2 * self.f

        return q



"""
Leimkuhler-Matthews (LM)
========================

A biased MCMC sampling method that provides significant accuracy improvements over the Euler-Maruyama scheme, without extra gradient evaluations.
Gives `exact sampling` for Gaussian target distributions.

Usage
^^^^^

Set ``algo="LM"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The gradient of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- This scheme takes no additional parameters.

References
^^^^^^^^^^

- `Molecular Dynamics: With Deterministic and Stochastic Numerical Methods by Benedict Leimkuhler and Charles Matthews, Springer (2017)`

"""
from .algorithm import Algorithm


class LM(Algorithm):
    def __init__(self, np, ic, h, force, params):
        super().__init__(np, ic, h, force, params)

        self.R = np.random.randn(*ic.shape)

    def step(self, q):

        q = q + self.h * self.f + self.sqh2 * self.R

        self.R = self.np.random.randn(*q.shape)

        q = q + self.sqh2 * self.R

        fres = self.force(q)
        self.v, self.f = fres.get("llh"), fres.get("grad")

        return q



"""
:: racecar sampler class

A class containing the state information for the current sampler object.

"""

from __future__ import absolute_import
import time
import numpy as np

from .algos.hmc import HMC
from .algos.lm import LM
from .algos.racecar import RACECAR
from .algos.sgld import SGLD
from .algos.badodab import BADODAB
from .algos.baoab import BAOAB
from .algos.rwmetropolis import RWMETROPOLIS
from .algos.sghmc import SGHMC


class Sampler:
    """
    A class containing the state information for the current sampler object.

    Parameters
    ----------
    ic : numpy array
        initial condition for the sampler
    h : float, positive
        the step size or learning rate to use in the sampler
    llh : function
        The log posterior function to use for the sampler. The function should take the position as input, and output a dictionary.
        The information required is algorithm-dependent. The algorithms make use of keys:

        - `"llh"` (float) : The value of the posterior function at any point
        - `"grad"` (numpy array) : The gradient of the posterior
        - `"grad_data"` (numpy array) : The gradients of the posterior for each data point

    algo : string, optional
        The algorithm to use for the sampler. See the :ref:`algorithms` page  for a list of possible methods supported by the package. Uses the `racecar` algorithm as default.
    params : dict, optional
        parameters
    """

    def __init__(self, ic, h, llh, algo="racecar", params={}):
        """
        Constructs all the necessary attributes for the sampler object.
        """

        self.timetaken_value = 0

        ic = np.array(ic).reshape(-1, 1)
        self.q = np.copy(ic)

        if params.get("seed") is not None:
            np.random.seed(params.get("seed"))

        args = (np, ic, h, llh, params)

        if algo.lower() == "sgld":
            self.ig = SGLD(*args)

        if algo.lower() == "lm":
            self.ig = LM(*args)

        if algo.lower() == "hmc":
            self.ig = HMC(*args)

        if algo.lower() == "racecar":
            self.ig = RACECAR(*args)

        if algo.lower() == "badodab":
            self.ig = BADODAB(*args)

        if algo.lower() == "baoab":
            self.ig = BAOAB(*args)

        if algo.lower() == "rwmetropolis":
            self.ig = RWMETROPOLIS(*args)

        if algo.lower() == "sghmc":
            self.ig = SGHMC(*args)

    def sample(self, Nsteps, printnum=None, thin=1, output=["pos"]):
        """
        Runs the sampler and returns a trajectory (if required).

        Parameters
        ----------
        Nsteps : int
            the number of steps to run.
        printnum : int, optional
            if given, prints a  summary of the sampling ``printnum`` times
        thin : int, optional
            thins the trajectory by only including one of each ``thin`` points
        output : list, optional
            a list of strings specifying which arrays should be returned by the function.
            The output list can contain

            - "pos" : Outputs the trajectory of position
            - "grad" : Outputs the gradient at each point
            - "llh" : Output the log posterior evaluated at each point
            - "xi" : Gives the auxilliary variable's value at each step
            - 'mom' : The sampled momentum points

            The default value is ``["pos"]``.

        Returns
        -------
        Q : numpy array
            the trajectory array for sampled points
        P : numpy array
            the sampled momentum points (if applicable)
        LLH : numpy array
            the log posterior of sampled points
        XI : numpy array
            the sampled auxillary variables (if applicable)
        F : numpy array
            the gradient of the log posterior function at each sampled point (if applicable)
        """

        Qk = 0
        NQ = 1 + Nsteps // thin
        self.ig.acc = 0
        self.ig.clear(self.q)

        output_dict = {}
        for o in output:
            output_dict[o.lower()] = True

        if output_dict.get("pos"):
            Q = np.zeros([NQ, len(self.q.flatten())])
        if output_dict.get("mom"):
            P = np.zeros([NQ, len(self.ig.p.flatten())])
        if output_dict.get("llh"):
            LLH = np.zeros([NQ, 1])
        if output_dict.get("grad"):
            F = np.zeros([NQ, len(self.ig.f.flatten())])
        if output_dict.get("xi"):
            XI = np.zeros([NQ, len(self.ig.xi.flatten())])

        stime = time.time()

        if printnum is not None:
            pf = Nsteps // printnum
        else:
            pf = Nsteps + 2

        for n in range(Nsteps):

            self.q = self.ig.step(self.q)

            if (n % thin == 0) and (Qk < NQ):
                if output_dict.get("pos"):
                    Q[Qk, :] = self.q.copy().flatten()
                if output_dict.get("mom"):
                    P[Qk, :] = self.ig.p.copy().flatten()
                if output_dict.get("llh"):
                    LLH[Qk] = self.ig.v
                if output_dict.get("grad"):
                    F[Qk, :] = self.ig.f.copy().flatten()
                if output_dict.get("xi"):
                    XI[Qk, :] = self.ig.xi.copy().flatten()

                Qk += 1

            if np.isnan(self.q).any():
                break

            if (n + 1) % pf == 0:
                print(
                 "Steps: %5d | Time: %7.3f | V: %+7.3e" % (n+1,time.time() - stime, self.ig.v )
                )

        self.timetaken_value = time.time() - stime
        self.Nsteps_value = Nsteps

        output_tuple = ()
        for o in output:
            if o.lower() == "pos":
                output_tuple += (Q[:Qk, :],)
            if o.lower() == "mom":
                output_tuple += (P[:Qk, :],)
            if o.lower() == "llh":
                output_tuple += (LLH[:Qk],)
            if o.lower() == "xi":
                output_tuple += (XI[:Qk, :],)
            if o.lower() == "grad":
                output_tuple += (F[:Qk, :],)

        if len(output_tuple) == 1:
            return output_tuple[0]

        return output_tuple

    def timetaken(self):
        """
        Returns the wall time of the previously run ``sample`` call.

        Returns
        -------
        T : float
            the wall time of the last sampling run, in seconds.
        """

        return self.timetaken_value

    def acceptance_rate(self):
        """
        Returns the acceptance rate of the previously run ``sample`` call.

        Returns
        -------
        A : float
            the acceptance rate of the last sampling run, if applicable.
        """

        if (self.ig.acc==0):
            return 0

        return self.ig.acc / self.Nsteps_value



"""Racecar

A lightweight Python sampling package.

See https://github.com/c-matthews/racecar for information and examples.

"""
from __future__ import absolute_import

# Set explicit packages
from .sampler import Sampler as sampler
from .llh import blr, isotropic_gaussian, ica
from ._version import version_number

__author__ = "Charles Matthews"
__license__ = "MIT"
__maintainer__ = "Charles Matthews"
__email__ = "mail@cmatthe.ws"
__version__ = version_number



"""
Random Walk Metropolis
======================

One of the simplest and widely used MCMC sampling algorithms, the Metropolis algorithm proposes a new point using an isotropic Gaussian distribution and accepts with a probability preserving the detailed balance condition.

The Metropolis condition ensures this method is unbiased, but can be extremely slow to converge (particularly in high dimension).

.. note::
    The acceptance rate for the scheme can be checked by calling the ``acceptance_rate`` function on the sampler. A good choice is to choose a learning rate so that acceptance is around 0.234.

Usage
^^^^^

Set ``algo="RWMetropolis"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``llh`` :: The value of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- This scheme takes no additional parameters.
"""
from .algorithm import Algorithm


class RWMETROPOLIS(Algorithm):
    def step(self, q):

        p = self.np.random.randn(*q.shape)

        q0 = q.copy()

        V0 = self.v

        h = self.h

        q = q + h * p
        fres = self.force(q)
        V1, F = fres.get("llh"), fres.get("grad")

        dH = V0 - V1

        if self.np.log(self.np.random.rand()) < -dH:
            # Accept
            self.acc += 1
            self.v = V1
            self.f = F
            return q

        # Not accept
        return q0



import setuptools
import src.racecar._version as v

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="racecar",
    version=v.version_number,
    author="Charles Matthews",
    author_email="mail@cmatthe.ws",
    description="A lightweight Bayesian sampling package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/c-matthews/racecar",
    install_requires=['numpy', 'scipy'],
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



"""
Racecar
========================

A general purpose sampling algorithm, specifically designed for efficient sampling in systems with a stochastic gradient.

Usage
^^^^^

Set ``algo="racecar"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``grad`` :: The gradient of the log posterior.
- ``grad_data`` :: (optional) The gradient of the log likelihood with respect to the current position, itemized over the data batch. Expects a `(N,D)` array, where the position space is `N` dimensional and the batch size is `D`.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- ``g`` : positive float
    (Default 1.0) The damping parameter used in the Langevin dynamics. Must be large enough to dominate the force noise.
- ``mu`` : positive float
    (Default 1.0) The precision (reciprocal variance) of the auxillary variables distribution.
- ``estimate_basis`` : bool
    (Default `True`) If set to true and ``grad_data`` is given, then the method will approximate the dominant directions for the gradient noise.
- ``estimate_time`` : float
    (Default 0.5) The time in seconds to spend initially estimating the basis vectors by generating new grad-data. By default it will spend 0.5s building this at setup. Set it to 0 to just use the initial `grad-data` for the guess.
- ``basis_size`` : positive int
    (Default `Ndim`) The number of basis vectors to use. A smaller number can improve efficiency in high dimensions, but will remove potential accuracy.
- ``basis`` : `NxK` numpy array
    (optional) The rank-K array of basis vectors to use for the damping of the noisy gradient.

.. note::
    The `Racecar` scheme damps the system in directions specified by basis vectors.
    If a ``basis`` is not given, and ``estimate_basis`` is True, then these vectors are estimated from the ``grad_data`` value if it is outputted from the ``llh`` function. If there is no ``grad_data``, then it will the standard (diagonal) basis instead.

"""
from .algorithm import Algorithm
import time


class RACECAR(Algorithm):
    def __init__(self, np, ic, h, force, params):

        super().__init__(np, ic, h, force, params)

        self.g = params.get("g", 1)
        self.mu = params.get("mu", 1)
        self.c1 = self.np.exp(-self.h * self.g)
        self.c3 = self.np.sqrt(1 - self.c1 * self.c1)
        ff = self.ff

        Ndim = ic.size

        self.use_basis = False
        if params.get("basis") is not None:
            self.B = params.get("basis")
            self.use_basis = True
        else:
            if (ff is not None) and (params.get("estimate_basis", True)):
                assert ff.shape[0] == Ndim
                my_grad_data = [ff]
                stime = time.time()
                while (time.time()-stime<=params.get("estimate_time",0.5)):
                    my_grad_data.append( self.force(ic)['grad_data'] )
                my_grad_data = np.hstack(my_grad_data)
                evals, self.B = np.linalg.eig((1e-6) * np.eye(Ndim) + np.cov(my_grad_data))
                evals = evals.real
                self.B = self.B.real

                sz = params.get("basis_size")
                if sz is not None:
                    self.B = self.B[:, self.np.argsort(evals)[-sz:]]
                self.use_basis = True

        if self.use_basis:
            self.xi = (1 / self.mu) * np.random.randn(self.B.shape[1], 1)
        else:
            self.xi = (1 / self.mu) * np.random.randn(Ndim, 1)

    def step_C(self):

        dh = [1.351207191959657, -1.702414383919315, 1.351207191959657]

        if self.use_basis:
            Bp = self.np.dot(self.B.T, self.p)
        else:
            Bp = self.p

        xi = self.xi

        Bp, xi = self.leapfrog_C(Bp, xi, dh[0] * self.h2)
        Bp, xi = self.leapfrog_C(Bp, xi, dh[1] * self.h2)
        Bp, xi = self.leapfrog_C(Bp, xi, dh[2] * self.h2)

        self.xi = xi

        if self.use_basis:
            self.p = self.p - self.np.linalg.multi_dot([self.B, self.B.T, self.p])
            self.p = self.p + self.np.dot(self.B, Bp)
        else:
            self.p = Bp

        return

    def leapfrog_C(self, pp, xx, h):

        h2 = h / 2

        xx = xx + (pp * pp - 1) * (h2 / self.mu)
        pp = pp * self.np.exp(-h * xx)
        xx = xx + (pp * pp - 1) * (h2 / self.mu)

        return pp, xx

    def step_E(self):

        self.p = self.c1 * self.p + self.c3 * self.np.random.randn(*self.p.shape)

        return

    def step(self, q):

        h = self.h
        h2 = self.h2

        # R
        self.p = self.p + h2 * self.f

        # A
        q = q + h2 * self.p

        # C
        self.step_C()

        # E
        self.step_E()

        # C
        self.step_C()

        # A
        q = q + h2 * self.p

        fres = self.force(q)
        self.v, self.f = fres.get("llh"), fres.get("grad")

        # R
        self.p = self.p + h2 * self.f

        return q



"""
Hybrid Monte Carlo (HMC)
========================

Hybrid (or Hamiltonian) Monte Carlo takes steps of constant-energy dynamics after redrawing the kinetic energy. The symplectic scheme used for the constant-energy integration allows for very favorable behavior with dimension, in spite of the Metropolization step.

.. note::
    The acceptance rate for the scheme can be checked by calling the ``acceptance_rate`` function on the sampler. A good choice is to choose a learning rate so that acceptance is between 0.7 and 0.8.

Usage
^^^^^

Set ``algo="hmc"`` in :ref:`sampler`.

Requires
^^^^^^^^

The ``llh`` function needs to output the following dictionary keys:

- ``llh`` :: The log posterior value.
- ``grad`` :: The gradient of the log posterior.

Params
^^^^^^

The behavior of the sampler can be customized by including the following arguments in the sampler's ``params`` dict.

- ``T`` : positive int
    (Default 5) The number of constant-energy substeps to take between Metropolis tests.

References
^^^^^^^^^^

- `Hybrid Monte Carlo; Simon Duane, A.D. Kennedy, Brian J. Pendleton, Duncan Roweth (1987)` `<https://doi.org/10.1016/0370-2693(87)91197-X>`__

"""
from .algorithm import Algorithm


class HMC(Algorithm):
    def __init__(self, np, ic, h, force, params):
        super().__init__(np, ic, h, force, params)

        self.T = params.get("T", 5)

    def step(self, q):

        p = self.np.random.randn(*q.shape)

        q0 = q.copy()

        V0 = self.v
        K0 = -self.np.sum(p * p) / 2

        f = self.f.copy()

        h2 = self.h2
        h = self.h

        for m in range(self.T):
            p = p + h2 * f
            q = q + h * p
            fres = self.force(q)
            V1, f = fres.get("llh"), fres.get("grad")
            p = p + h2 * f

        K1 = -self.np.sum(p * p) / 2

        dH = (V0 - V1) + (K0 - K1)

        if self.np.log(self.np.random.rand()) < -dH:
            # Accept
            self.acc += 1
            self.v = V1
            self.f = f.copy()
            self.p = p.copy()
            return q

        # Not accept
        return q0
